// MessageTextView version 1.0
// by Al.V.Sarikov.
// Kherson, Ukraine, 2006.
// E-mail: avix@ukrpost.net.
// Home page: http://avix.pp.ru.

// This class provides an underlined and clickable BStringView.
// Being clicked it sends message the same as BButton control. 

// Class is created on base of and using code of URLView class of version 2.11
// written by William Kakes of Tall Hill Software (http://www.tallhill.com).

#include "MessageTextView.h"

#include <Application.h>
#include <Window.h>

// This is the link's mouse cursor (a replica of NetPositive's link cursor).
const uint8 url_cursor[]=
{
 16, 1, 1, 2,
 // This is the cursor data.
 0x00, 0x00, 0x38, 0x00, 0x24, 0x00, 0x24, 0x00,
 0x13, 0xe0, 0x12, 0x5c, 0x09, 0x2a, 0x08, 0x01,
 0x3c, 0x21, 0x4c, 0x71, 0x42, 0x71, 0x30, 0xf9,
 0x0c, 0xf9, 0x02, 0x00, 0x01, 0x00, 0x00, 0x00,
 
 // This is the cursor mask.
 0x00, 0x00, 0x38, 0x00, 0x3c, 0x00, 0x3c, 0x00,
 0x1f, 0xe0, 0x1f, 0xfc, 0x0f, 0xfe, 0x0f, 0xff,
 0x3f, 0xff, 0x7f, 0xff, 0x7f, 0xff, 0x3f, 0xff,
 0x0f, 0xff, 0x03, 0xfe, 0x01, 0xf8, 0x00, 0x00,
};

// The default link color, blue.
const rgb_color blue={0, 0, 255};
// The default clicked-link color, red.
const rgb_color red={255, 0, 0};
// The default link hover color, dark blue.
const rgb_color dark_blue={0, 0, 120};
// The default disabled color, gray.
const rgb_color gray={100, 100, 100};


//////////////////////////////////////////////////////////////////////////////////
MessageTextView::MessageTextView(BRect frame, const char *name, const char *label,
                                 BMessage *msg, uint32 resizingMode, uint32 flags)
                :BStringView(frame, name, label, resizingMode, flags),
                 BInvoker(msg, NULL)
{
 // Set the default values for the other definable instance variables.
 this->color=blue;
 this->clickColor=red;
 this->hoverColor=dark_blue;
 this->disabledColor=gray;
 this->hoverEnabled=true;
 this->underlineThickness=1;
 
 // The link should be enabled by default.
 enabled=true;
 
 // Create the cursor to use when over the link.
 this->linkCursor=new BCursor(url_cursor);
 
 // The link is not currently selected.
 selected=false;
 
 // The URL is currently not hover-colored.
 hovering=false;
}
  

///////////////////////////////////
MessageTextView::~MessageTextView()
{
 delete linkCursor;
}


////////////////////////////////////////
void MessageTextView::AttachedToWindow()
{
 // When the view is first attached, we want to draw the link
 // in the normal color.  Also, we want to set our background color
 // to meet that of our parent.
 if(IsEnabled()) SetHighColor(color);
 else SetHighColor(disabledColor);
  
 if(Parent()!=NULL)
 {
  SetLowColor(Parent()->ViewColor());
  SetViewColor(Parent()->ViewColor());
 }
 
 // Setting the target for message
 if(!(Messenger().IsValid())) if(Window()!=NULL) SetTarget(Window());
}


////////////////////////////////////////////
void MessageTextView::Draw(BRect updateRect)
{
 BRect rect=Frame();
 rect.OffsetTo(B_ORIGIN);
 
 // We want 'g's, etc. to go below the underline.  When the BeOS can
 // do underlining of any font, this code can be removed.
 font_height height;
 GetFontHeight(&height);
 float descent=height.descent;
 
 // We want to be sensitive to the SetAlignment() function.
 float left, right;
 if(Alignment()==B_ALIGN_RIGHT)
 {
  left=rect.right-StringWidth(Text())+1;
  right=rect.right-1;
 }
 else if(Alignment()==B_ALIGN_CENTER)
 {
  left=rect.left+(rect.Width()/2)-(StringWidth(Text())/2);
  right=left+StringWidth(Text())-2;
 }
 else
 {
  left=rect.left;
  right=left+StringWidth(Text())-2;
 }
 
 // Draw the underline in the requested thickness.
 // Note:  If the link is disabled, we don't want to draw the underline.
 if(IsEnabled())
 {
  FillRect(BRect((float) left, (float) (rect.bottom-descent),
                 (float) right,(float) (rect.bottom-descent+(underlineThickness-1))));
 }
 MovePenTo(BPoint(left, rect.bottom-descent-1));
 
 // Note: DrawString() draws the text at one pixel above the pen's
 // current y coordinate.
 DrawString(Text());
}


/////////////////////////////////////////////
void MessageTextView::MouseDown(BPoint point)
{
 // If the link isn't enabled, don't do anything.
 if(!IsEnabled()) return;
 
 // See which mouse buttons were clicked.
// int32 buttons=Window()->CurrentMessage()->FindInt32("buttons");
 
 // We want to highlight the text if the user clicks on
 // the MessageTextView. We want to be sure to only register a click
 // if the user clicks on the link text itself and not just
 // anywhere in the view.
 if(GetTextRect().Contains(point))
 {
  SetHighColor(clickColor);
  Redraw();
  
  // Set the link as selected and track the mouse.
  selected=true;
  SetMouseEventMask(B_POINTER_EVENTS);
  
//  if(buttons==B_SECONDARY_MOUSE_BUTTON) inPopup=true;
 }
}


//////////////////////////////////////////////////////////////
void MessageTextView::MouseMoved(BPoint point, uint32 transit,
                                 const BMessage *message)
{
 // If the link isn't enabled, don't do anything.      
 if(!IsEnabled()) return;
 
 // Make sure the window is the active one.
 if(!Window()->IsActive()) return;
 
 // See which mouse buttons were clicked.
 int32 buttons=Window()->CurrentMessage()->FindInt32("buttons");
 
 switch(transit)
 {
  case(B_ENTERED_VIEW):
   // Should we set the cursor to the link cursor?
   if(GetTextRect().Contains(point))
   {
    be_app->SetCursor(linkCursor);
    
    // Did the user leave and re-enter the view while
    // holding down the mouse button?  If so, highlight
    // the link.
    if(selected)
    {
     SetHighColor(clickColor);
     Redraw();
    }
    // Should we hover-highlight the link?
    else if(hoverEnabled)
    {
     if(buttons==0)
     {
      SetHighColor(hoverColor);
      Redraw();
      hovering=true;
     }
    } 
   }
   break;
  
  case(B_EXITED_VIEW):
   // We want to restore the link to it normal color and the
   // mouse cursor to the normal hand.  However, we should only
   // set the color and re-draw if it is needed.
   if(selected)
   {
    be_app->SetCursor(B_HAND_CURSOR);
    SetHighColor(color);
    Redraw();
   }
   // Is the link currently hover-highlighted?  If so, restore
   // the normal color now.
   else if(hovering)
   {
    be_app->SetCursor(B_HAND_CURSOR);
    SetHighColor(color);
    Redraw();
    hovering=false;
   }
   // Change the cursor back to the hand.
   else
   {
    be_app->SetCursor(B_HAND_CURSOR);
   }
   break;
  
  case(B_INSIDE_VIEW):
   // The user could either be moving out of the view or
   // back into it here, so we must handle both cases.
   // In the first case, the cursor is now over the link.
   if(GetTextRect().Contains(point))
   {
    // We only want to change the cursor.      
    be_app->SetCursor(linkCursor);
    if(selected)
    {
     // Highlight link as long as the user holds the button down
     // and has the mouse cursor over it (like a standard button).
     SetHighColor(clickColor);
     Redraw();
    }
    // The link isn't currently selected?  If hover-highlighting
    // is enabled, highlight the link.
    else if(hoverEnabled)
    {
     SetHighColor(hoverColor);
     Redraw();
     hovering=true;
    }
   }
   // In this case, the mouse cursor is not over the link, so we
   // need to restore the original link color, etc.
   else
   {
    be_app->SetCursor(B_HAND_CURSOR);
    if(selected)
    {
     SetHighColor(color);
     Redraw();
    }
    // Is the mouse cursor hovering over the link?
    else if(hovering)
    {
     SetHighColor(color);
     Redraw();
     hovering=false;
    }
   }
   break;
 }
}


///////////////////////////////////////////
void MessageTextView::MouseUp(BPoint point)
{
 // If the link isn't enabled, don't do anything.
 if(!IsEnabled()) return;
 
 // If the link was clicked on, send message to the target.
 if(GetTextRect().Contains(point) &&  hoverEnabled) SetHighColor(hoverColor);
 else if(!hovering) SetHighColor(color);
 Redraw();
 
 if(selected  &&  GetTextRect().Contains(point)) InvokeNotify(Message());
 
 selected=false;
 
 // Should we restore the hovering-highlighted color or the original
 // link color?
}


//////////////////////////////////////////////////
void MessageTextView::WindowActivated(bool active)
{
 // Be sure that if the user clicks on a link prompting the opening of
 // a new window (i.e. launching NetPositive) the URL is not left drawn
 // with the hover color.
 if(!active)
 {
  if(IsEnabled())
  {
   SetHighColor(color);
   Redraw();
  }
 }
}


/////////////////////////////////
bool MessageTextView::IsEnabled()
{
 // Return whether or not this link is enabled (and therefore clickable).
 return enabled;
}


////////////////////////////////////////
rgb_color MessageTextView::Color() const
{
 // Returns normal link color.
 return color;
}


///////////////////////////////////////////////
void MessageTextView::SetColor(rgb_color color)
{
 // Set the normal link color.
 this->color=color;
 if(IsEnabled())
 {
  Window()->Lock();
  SetHighColor(color);
  Redraw();
  Window()->Unlock();
 }
}


///////////////////////////////////////////////////////////////////////////////
void MessageTextView::SetColor(uchar red, uchar green, uchar blue, uchar alpha)
{
 // Set the normal link color.
 rgb_color color;
 color.red=red;
 color.green=green;
 color.blue=blue;
 color.alpha=alpha;
 SetColor(color);
}


/////////////////////////////////////////////
rgb_color MessageTextView::ClickColor() const
{
 // Returns link color used when the link is clicked.
 return clickColor;
}


////////////////////////////////////////////////////
void MessageTextView::SetClickColor(rgb_color color)
{
 // Set the link color used when the link is clicked.
 clickColor=color;
}


////////////////////////////////////////////////////////////////////////////////////
void MessageTextView::SetClickColor(uchar red, uchar green, uchar blue, uchar alpha)
{
 // Set the link color used when the link is clicked.
 rgb_color color;
 color.red=red;
 color.green=green;
 color.blue=blue;
 color.alpha=alpha;
 SetClickColor(color);
}


////////////////////////////////////////////////
rgb_color MessageTextView::DisabledColor() const
{
 // Returns color of disabled link.
 return disabledColor;
}


///////////////////////////////////////////////////////
void MessageTextView::SetDisabledColor(rgb_color color)
{
 // Set the color to draw in when the link is disabled.
 disabledColor=color;
 Window()->Lock();
 Redraw();
 Window()->Unlock();
}


///////////////////////////////////////////////////////////////////////////////////////
void MessageTextView::SetDisabledColor(uchar red, uchar green, uchar blue, uchar alpha)
{
 // Set the color to draw in when the link is disabled.
 rgb_color color;
 color.red=red;
 color.green=green;
 color.blue=blue;
 color.alpha=alpha;
 SetDisabledColor(color);
}


//////////////////////////////////////////////
void MessageTextView::SetEnabled(bool enabled)
{
 // Set whether or not the link is enabled (and therefore clickable).
 bool redraw=this->enabled!=enabled;
 this->enabled=enabled;
 
 if(Window())
 {
  Window()->Lock();
  if(!enabled) SetHighColor(disabledColor);
  else SetHighColor(color);
  if(redraw) Invalidate();
  Window()->Unlock();
 }
}


/////////////////////////////////////////////
rgb_color MessageTextView::HoverColor() const
{
 // Returns link color used when the mouse cursor is over it.
 return hoverColor;
}


////////////////////////////////////////////////////
void MessageTextView::SetHoverColor(rgb_color color)
{
 // Set the link color used when the mouse cursor is over it.
 hoverColor=color;
}


////////////////////////////////////////////////////////////////////////////////////
void MessageTextView::SetHoverColor(uchar red, uchar green, uchar blue, uchar alpha)
{
 // Set the color to draw in when the link is disabled.
 rgb_color color;
 color.red=red;
 color.green=green;
 color.blue=blue;
 color.alpha=alpha;
 SetHoverColor(color);
}


//////////////////////////////////////
bool MessageTextView::IsHoverEnabled()
{
 // Returns whether or not to hover-highlight the link.
 return hoverEnabled;
}


/////////////////////////////////////////////////
void MessageTextView::SetHoverEnabled(bool hover)
{
 // Set whether or not to hover-highlight the link.
 hoverEnabled=hover;
}


/////////////////////////////////////////
int MessageTextView::UnderlineThickness()
{
 // Returns the thickness of the underline in pixels.
 return underlineThickness;
}


//////////////////////////////////////////////////////////
void MessageTextView::SetUnderlineThickness(int thickness)
{
 // Set the thickness of the underline in pixels.
 underlineThickness=thickness;
}


////////////////////////////////////
BRect MessageTextView::GetTextRect()
{
 // This function will return a BRect that contains only the text
 // and the underline, so the mouse can change and the link will
 // be activated only when the mouse is over the text itself, not
 // just within the view.
 
 // Note:  We'll use bounding boxes, because they are the most
 //        accurate, and since the user is interacting with the
 //        view, off-by-one-pixel errors look bad.
 const char *textArray[1];
 textArray[0]=Text();
 
 escapement_delta delta;
 delta.nonspace=0;
 delta.space=0;
 escapement_delta escapements[1];
 escapements[0]=delta;
 
 BRect returnMe;
 BRect rectArray[1];
 rectArray[0]=returnMe;
 
 BFont font;
 GetFont(&font);
 font.GetBoundingBoxesForStrings(textArray, 1, B_SCREEN_METRIC, escapements, rectArray);

 BRect frame=Frame();
 frame.OffsetTo(B_ORIGIN);
 returnMe=rectArray[0];
 
 // Get the height of the current font.
 font_height height;
 GetFontHeight(&height);
 float descent=height.descent;
 
 // Account for rounding of the floats when drawn to avoid
 // one-pixel-off errors.
 float lowerBound=0;
 if((((int) descent)*2)!=((int) (descent*2))) lowerBound=1;
 
 // Adjust the bounding box to reflect where the text is in the view.
 returnMe.bottom+=1;
 returnMe.OffsetTo(B_ORIGIN);
 float rectHeight=returnMe.Height();
 returnMe.bottom=frame.bottom-descent;
 returnMe.top=returnMe.bottom-rectHeight;
 returnMe.bottom+=1+underlineThickness;
 returnMe.OffsetBy(0.0, -(1+lowerBound));

 return returnMe;
}


//////////////////////////////
void MessageTextView::Redraw()
{
 // Redraw the link without flicker.
 BRect frame=Frame();
 frame.OffsetTo(B_ORIGIN);
 Draw(frame);
}
